
@implementation StatusPackageModel

- (NSArray <InputPackageFile *>*)files {
    NSString *baseFile = [NSString stringWithFormat:@"%@.list", self.package];
    NSString *listFile = [self.bootstrapPath.relativeInfoFolderPath stringByAppendingPathComponent:baseFile];
    NSArray *files = [[NSString stringWithContentsOfFile:listFile encoding:NSUTF8StringEncoding error:nil] componentsSeparatedByString:@"\n"];
    __block NSMutableArray *newFiles = [NSMutableArray new];
    [files enumerateObjectsUsingBlock:^(id  _Nonnull obj, NSUInteger idx, BOOL * _Nonnull stop) {
        BOOL isDir = false;
        NSString *fullPath = [self.bootstrapPath stringByAppendingPathComponent:obj];
        if ([FM fileExistsAtPath:fullPath isDirectory:&isDir]){
                InputPackageFile *file = [[InputPackageFile alloc] initWithFile:obj inBootstrap:self.bootstrapPath];
                if (file) {
                    [newFiles addObject:file];
                }
        } else { //if it doesn't exist its probably a symbolic link that doensn't point to any 'proper' location on our current device.
            InputPackageFile *file = [[InputPackageFile alloc] initWithFile:obj inBootstrap:self.bootstrapPath];
            if (file) {
                [newFiles addObject:file];
            }
        }
    }];
    
    return newFiles;
}

- (NSArray *)infoFiles {
    NSArray *infos = [FM contentsOfDirectoryAtPath:self.bootstrapPath.relativeInfoFolderPath error:nil];
    return [infos filteredArrayUsingPredicate:[NSPredicate predicateWithFormat:@"SELF BEGINSWITH %@", self.package]];
}

- (NSString *)writeableControl {
    NSMutableArray *controlArray = [[self.rawString componentsSeparatedByString:@"\n"] mutableCopy];
    [controlArray removeLastObject];
    [controlArray addObject:@"\n"];
    return [controlArray componentsJoinedByString:@"\n"];
}

- (void)createPackage {
    NSArray *infoFiles = [self infoFiles];
    NSArray <InputPackageFile *> *files = [self files];
    NSString *tmpFolder = [NSString stringWithFormat:@"/tmp/%@", self.package];
    if([FM createDirectoryAtPath:tmpFolder withIntermediateDirectories:true attributes:nil error:nil]) {
        DLog(@"directory created successfully: %@", tmpFolder);
        NSString *debFolderPath = [tmpFolder stringByAppendingPathComponent:@"DEBIAN"];
        [FM createDirectoryAtPath:debFolderPath withIntermediateDirectories:true attributes:nil error:nil];
        NSString *control = [debFolderPath stringByAppendingPathComponent:@"control"];
        [self.writeableControl writeToFile:control atomically:true encoding:NSUTF8StringEncoding error:nil];
        [infoFiles enumerateObjectsUsingBlock:^(id  _Nonnull obj, NSUInteger idx, BOOL * _Nonnull stop) {
            if (![[obj pathExtension] isEqualToString:@"list"] && ![[obj pathExtension] isEqualToString:@"keep"]){
                NSString *fullPath = [self.bootstrapPath.relativeInfoFolderPath stringByAppendingPathComponent:obj];
                NSString *finalPath = [debFolderPath stringByAppendingPathComponent:[obj pathExtension]];
                [FM copyItemAtPath:fullPath toPath:finalPath error:nil];
            }
        }];
        [files enumerateObjectsUsingBlock:^(InputPackageFile * _Nonnull file, NSUInteger fileIdx, BOOL * _Nonnull stop) {
            NSString *destinationPath = [tmpFolder stringByAppendingPathComponent:file.path];
            if (file.type == BSPackageFileTypeDirectory) {
                [FM createDirectoryAtPath:destinationPath withIntermediateDirectories:true attributes:nil error:nil];
            } else {
                NSString *fromFile = [self.bootstrapPath stringByAppendingPathComponent:file.path];
                DLog(@"copying: %@ to %@", fromFile, destinationPath);
                [FM copyItemAtPath:fromFile toPath:destinationPath error:nil];
            }
        }];
    }
}

- (NSString *)description {
    NSString *orig = [super description];
    return [NSString stringWithFormat:@"%@ = %@ (%@)", orig, self.package, self.version];
}

- (NSString*) fullDescription {
    NSString *orig = [super description];
    NSMutableDictionary *details = [NSMutableDictionary new];
    NSArray *props = [self properties];
    [props enumerateObjectsUsingBlock:^(id  _Nonnull obj, NSUInteger idx, BOOL * _Nonnull stop) {
        NSString *cv = [self valueForKey:obj];
        if (cv){
            details[obj] = cv;
        }
    }];
    return [NSString stringWithFormat:@"%@ = %@", orig, details];
    
}

+ (NSDictionary *)dependencyDictionaryFromString:(NSString *)depend {
    NSMutableCharacterSet *whitespaceAndPunctuationSet = [NSMutableCharacterSet characterSetWithCharactersInString:@"()"];
    [whitespaceAndPunctuationSet formUnionWithCharacterSet:[NSCharacterSet whitespaceAndNewlineCharacterSet]];
    NSScanner *stringScanner = [[NSScanner alloc] initWithString:depend];
    stringScanner.charactersToBeSkipped = whitespaceAndPunctuationSet;
    NSString *name = nil;
    NSInteger i = 0;
    NSMutableDictionary *predicate = [NSMutableDictionary new];
    while ([stringScanner scanUpToCharactersFromSet:whitespaceAndPunctuationSet intoString:&name]) {
        // DLog(@"%@ pass %li", name, (long)i);
        switch (i) {
            case 0 :
                predicate[@"package"] = name;
                break;
            case 1:
                predicate[@"predicate"] = name;
                break;
            case 2:
                predicate[@"requirement"] = name;
                break;
            default:
                break;
        }
        i++;
    }
    return predicate;
}

+ (NSArray *)dependencyArrayFromString:(NSString *)depends {
    NSMutableArray *cleanArray = [[NSMutableArray alloc] init];
    NSArray *dependsArray = [depends componentsSeparatedByString:@","];
    for (id depend in dependsArray) {
        NSArray *spaceDelimitedArray = [depend componentsSeparatedByString:@" "];
        if (spaceDelimitedArray.count > 1){
            NSString *isolatedDependency = [[spaceDelimitedArray objectAtIndex:0] stringByTrimmingCharactersInSet:[NSCharacterSet whitespaceAndNewlineCharacterSet]];
            if ([isolatedDependency length] == 0)
                isolatedDependency = [[spaceDelimitedArray objectAtIndex:1] stringByTrimmingCharactersInSet:[NSCharacterSet whitespaceAndNewlineCharacterSet]];
            
            NSDictionary *dependDict = [self dependencyDictionaryFromString:depend];
            //DLog(@"depend dict: %@", dependDict);
            [cleanArray addObject:dependDict];
        }
    }
    return cleanArray;
}

+ (NSDictionary *)JSONKeyPathsByPropertyKey {
    return @{
             @"name": @"Name",
             @"package": @"Package",
             @"source": @"Source",
             @"version": @"Version",
             @"priority": @"Priority",
             @"essential": @"Essential",
             @"depends": @"Depends",
             @"maintainer": @"Maintainer",
             @"packageDescription": @"Description",
             @"homepage": @"Homepage",
             @"icon": @"Icon",
             @"author": @"Author",
             @"preDepends": @"Pre-Depends",
             @"breaks": @"Breaks",
             @"depiction": @"Depiction",
             @"tag": @"Tag",
             @"architecture": @"Architecture",
             @"section": @"Section",
             @"osMax": @"osMax",
             @"osMin": @"osMin",
             @"banner": @"Banner",
             @"topShelfImage": @"TopShelfImage"
             };
}

- (instancetype)initWithRawControlString:(NSString *)controlString {
    NSArray *packageArray = [controlString componentsSeparatedByString:@"\n"];
    NSMutableDictionary *currentPackage = [[NSMutableDictionary alloc] init];
    for (id currentLine in packageArray) {
        NSArray *itemArray = [currentLine componentsSeparatedByString:@": "];
        if ([itemArray count] >= 2) {
            NSString *key = [itemArray objectAtIndex:0];
            NSString *object = [itemArray objectAtIndex:1];
            //process the array
            if ([key isEqualToString:@"Depends"]) {
                NSArray *dependsObject = [StatusPackageModel dependencyArrayFromString:object];
                [currentPackage setObject:dependsObject forKey:key];
            } else { //every other key, even if it has an array is treated as a string
                [currentPackage setObject:object forKey:key];
            }
        }
    }
    if ([[currentPackage allKeys] count] > 4) {
        self = [super init];
        self.rawString = controlString;
        [self mapDictionaryToProperties:currentPackage];
        return self;
    }
    return nil;
}

- (void)mapDictionaryToProperties:(NSDictionary *)theProps {
    NSArray *ourProps = [self properties];
    NSArray *allKeys = [theProps allKeys];
    NSDictionary *mappedKeys = [StatusPackageModel JSONKeyPathsByPropertyKey];
    [ourProps enumerateObjectsUsingBlock:^(id  _Nonnull obj, NSUInteger idx, BOOL * _Nonnull stop) {
        NSString *mappedProp = mappedKeys[obj];
        //DLog(@"allkeys: %@", allKeys);
        if ([allKeys containsObject:mappedProp]){
            if ([self respondsToSelector:NSSelectorFromString(obj)]){ //redudant
                id value = theProps[mappedProp];
                //DLog(@"setting value: %@ for key: %@ from mapped key: %@", value, obj, mappedProp);
                [self setValue:value forKey:obj];
            }
        } else {
            //DLog(@"%@ doesnt respond to %@", self, mappedProp);
        }
    }];
}

@end
